@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # -*- coding: latin-1 -*-
 # Date Format: YYYY-MM-DD
 
@@ -27,21 +27,25 @@ https://github.com/TeamMsgExtractor/msg-extractor
 #    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 __author__ = 'Destiny Peterson & Matthew Walker'
-__date__ = '2022-06-14'
-__version__ = '0.34.2'
+__date__ = '2022-07-11'
+__version__ = '0.35.1'
 
 import logging
 
 from . import constants, enums
-from .appointment import Appointment
+from .appointment import AppointmentMeeting
 from .attachment import Attachment
 from .contact import Contact
 from .exceptions import UnrecognizedMSGTypeError
+from .meeting_forward import MeetingForwardNotification
+from .meeting_request import MeetingRequest
+from .meeting_response import MeetingResponse
 from .message import Message
 from .message_base import MessageBase
 from .message_signed import MessageSigned
 from .message_signed_base import MessageSignedBase
 from .msg import MSGFile
+from .post import Post
 from .prop import createProp
 from .properties import Properties
 from .recipient import Recipient
